/*******************************************************************************
 * Copyright (c) 2009 Information-technology Promotion Agency, Japan.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package benten.cat.ui;

import org.eclipse.core.runtime.preferences.AbstractPreferenceInitializer;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Group;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPreferencePage;

import benten.cat.glossary.ui.CatGlossaryUiPlugin;
import benten.cat.tm.ui.CatTmUiPlugin;
import benten.cat.ui.CatUiPlugin.CatUiPreference;
import benten.cat.ui.messages.CatPreferencePageMessages;
import benten.ui.fields.TextField;
import benten.ui.preference.BentenPreference;
import benten.ui.views.UiToolkit;

/**
 * CAT プラグイン設定ページ。
 * 
 * @author KASHIHARA Shinji
 */
public class CatPreferencePage extends PreferencePage implements IWorkbenchPreferencePage {
	/**
	 * CAT プラグイン設定ページのためのメッセージ。
	 */
	protected static final CatPreferencePageMessages fMsg = new CatPreferencePageMessages();

	/**
	 * Benten が提供する基本的なウィジェットを生成する UI ツールキットのインスタンス。
	 */
	private final UiToolkit toolkit = new UiToolkit();

	// 画面項目。

	/** 翻訳の基軸言語・テキスト */
	private TextField transPivotLangText;

	/** 翻訳元の言語・テキスト */
	private TextField transSourceLangText;

	/** 翻訳先の言語・テキスト */
	private TextField transTargetLangText;

	/**
	 * 「改行、タブ、半角空白を可視化」ボタン。
	 */
	private Button showLineBreakCheck;

	@Override
	protected Control createContents(final Composite parent) {
		final Composite composite = toolkit.createComposite(parent, 1);

		final IPreferenceStore store = CatUiPlugin.getDefault().getPreferenceStore();
		final IPreferenceStore bentenStore = BentenPreference.getStore();

		{
			final Group transGroup = new Group(composite, SWT.BOLD);
			transGroup.setLayout(toolkit.createGridLayout(2, 5));
			transGroup.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
			transGroup.setText(fMsg.getLabelGroupTransLang());

			{
				String pivotLang = bentenStore.getString(BentenPreference.Preference.TRANS_PIVOT_LANG.name());
				if (pivotLang == null || pivotLang.length() == 0) {
					pivotLang = fMsg.getDefaultTransPivotLang();
				}

				transPivotLangText = new TextField(transGroup, fMsg.getLabelPivotLang());
				transPivotLangText.setWidthHint(SWT.DEFAULT);
				transPivotLangText.setText(pivotLang);
			}

			{
				String lang = bentenStore.getString(BentenPreference.Preference.TRANS_SOURCE_LANG.name());
				if (lang == null || lang.length() == 0) {
					lang = fMsg.getDefaultTransSourceLang();
				}

				transSourceLangText = new TextField(transGroup, fMsg.getLabelSourceLang());
				transSourceLangText.setWidthHint(SWT.DEFAULT);
				transSourceLangText.setText(lang);
			}

			{
				String lang = bentenStore.getString(BentenPreference.Preference.TRANS_TARGET_LANG.name());
				if (lang == null || lang.length() == 0) {
					lang = fMsg.getDefaultTransTargetLang();
				}

				transTargetLangText = new TextField(transGroup, fMsg.getLabelTargetLang());
				transTargetLangText.setWidthHint(SWT.DEFAULT);
				transTargetLangText.setText(lang);
			}

			toolkit.createLabel(composite, "☆上記の値は直接編集しないでください。これは製造途中の過渡期のものです。");
		}

		{
			final Group g = new Group(composite, SWT.BOLD);
			g.setLayout(toolkit.createGridLayout(1, 5));
			g.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
			g.setText(fMsg.getLabel1());

			showLineBreakCheck = new Button(g, SWT.CHECK);
			showLineBreakCheck.setText(fMsg.getLabel2());
			showLineBreakCheck.setSelection(store.getBoolean(CatUiPreference.SHOW_LINE_BREAK.name()));
		}

		return composite;
	}

	/**
	 * {@inheritDoc}
	 */
	public void init(final IWorkbench workbench) {
	}

	@Override
	public boolean performOk() {
		final IPreferenceStore store = CatUiPlugin.getDefault().getPreferenceStore();
		final IPreferenceStore bentenStore = BentenPreference.getStore();

		bentenStore.setValue(BentenPreference.Preference.TRANS_PIVOT_LANG.name(), transPivotLangText.getText());
		bentenStore.setValue(BentenPreference.Preference.TRANS_SOURCE_LANG.name(), transSourceLangText.getText());
		bentenStore.setValue(BentenPreference.Preference.TRANS_TARGET_LANG.name(), transTargetLangText.getText());

		store.setValue(CatUiPreference.SHOW_LINE_BREAK.name(), showLineBreakCheck.getSelection());

		CatUiPlugin.getDefault().firePreferenceChange();

		// 影響がある TM および用語集をリロードします。
		// TODO ここでは単にキャッシュクリアするだけのほうが良いかも知れません。
		CatTmUiPlugin.getDefault().loadExtensions();
		CatGlossaryUiPlugin.getDefault().loadExtensions();

		return true;
	}

	@Override
	protected void performDefaults() {
		final IPreferenceStore store = CatUiPlugin.getDefault().getPreferenceStore();

		transPivotLangText.setText(fMsg.getDefaultTransPivotLang());
		transSourceLangText.setText(fMsg.getDefaultTransSourceLang());
		transTargetLangText.setText(fMsg.getDefaultTransTargetLang());

		showLineBreakCheck.setSelection(store.getDefaultBoolean(CatUiPreference.SHOW_LINE_BREAK.name()));
	}

	/**
	 * 設定初期化クラス。
	 */
	public static class Initializer extends AbstractPreferenceInitializer {
		@Override
		public void initializeDefaultPreferences() {
			final IPreferenceStore store = CatUiPlugin.getDefault().getPreferenceStore();

			store.setDefault(CatUiPreference.SHOW_LINE_BREAK.name(), true);
		}
	}
}
